use axum::{
    async_trait,
    extract::{FromRef, FromRequest},
    http::{Request, StatusCode},
};
use ed25519_dalek::{Signature, VerifyingKey};
use hyper::Body;
use twilight_model::application::interaction::Interaction;

pub struct ExtractInteraction(pub Interaction);

#[async_trait]
impl<S> FromRequest<S, Body> for ExtractInteraction
where
    S: Send + Sync,
    VerifyingKey: FromRef<S>,
{
    type Rejection = (StatusCode, String);

    async fn from_request(request: Request<Body>, state: &S) -> Result<Self, Self::Rejection> {
        let (parts, body) = request.into_parts();
        let body_bytes = hyper::body::to_bytes(body)
            .await
            .map_err(|err| (StatusCode::INTERNAL_SERVER_ERROR, err.to_string()))?;

        let Ok(interaction): Result<Interaction, _> = serde_json::from_slice(&body_bytes) else {
            return Err((StatusCode::BAD_REQUEST, "request contained invalid json".to_string()))
        };
        let Some(sig) = parts.headers.get("x-signature-ed25519") else {
            return Err((StatusCode::BAD_REQUEST, "request did not include signature header".to_string()))
        };
        let Ok(sig) = hex::decode(sig) else {
            return Err((StatusCode::BAD_REQUEST, "requrest signature is invalid hex".to_string()))
        };
        let Ok(sig) = Signature::from_slice(&sig) else {
            return Err((StatusCode::BAD_REQUEST, "request signature is malformed".to_string()))
        };
        let Some(signed_buf) = parts.headers.get("x-signature-timestamp") else {
            return Err((StatusCode::BAD_REQUEST, "requrest did not include signature timestamp header".to_string()))
        };
        let mut signed_buf = signed_buf.as_bytes().to_owned();
        signed_buf.extend(body_bytes);

        let public_key = VerifyingKey::from_ref(state);
        let Ok(()) = public_key.verify_strict(&signed_buf, &sig) else {
            return Err((StatusCode::UNAUTHORIZED, "interaction failed signature verification".to_string()))
        };

        Ok(Self(interaction))
    }
}

