use std::{net::SocketAddr, process, str::FromStr};

use anyhow::bail;
use axum::{
    extract::{FromRef, State},
    http::StatusCode,
    routing::post,
    Json, Router,
};
use base64::{alphabet, engine, Engine};

use discord::interactions::ExtractInteraction;
use ed25519_dalek::VerifyingKey;

use sqlx::{postgres::PgPoolOptions, PgPool};
use tokio::signal;

use tower_http::trace::TraceLayer;
use tracing_subscriber::{
    filter::LevelFilter, layer::SubscriberExt, util::SubscriberInitExt, Layer,
};

use twilight_interactions::command::{CommandInputData, CommandModel, CreateCommand};

use twilight_model::{
    application::interaction::{InteractionData, InteractionType},
    http::interaction::{InteractionResponse, InteractionResponseType},
    id::{marker::ApplicationMarker, Id},
};

use crate::discord::commands::{register_commands, SetFactCommand, set_fact, GetFactCommand, get_fact};

mod database;
mod discord;

#[derive(Clone)]
struct Config {
    discord_client_id: Id<ApplicationMarker>,
    discord_client_secret: String,
    discord_pub_key: VerifyingKey,
    database_url: String,
    listen_port: u16,
}

impl Config {
    fn configure() -> anyhow::Result<Self> {
        dotenvy::dotenv().ok();

        let pub_key_bytes: Vec<u8> = hex::decode(std::env::var("DISCORD_PUB_KEY")?)?;
        let pub_key: [u8; 32] = match pub_key_bytes.try_into() {
            Ok(pk) => pk,
            Err(_) => bail!("Invalid discord public key"),
        };

        Ok(Config {
            discord_client_id: Id::from_str(std::env::var("DISCORD_CLIENT_ID")?.as_str())?,
            discord_client_secret: std::env::var("DISCORD_CLIENT_SECRET")?,
            discord_pub_key: VerifyingKey::from_bytes(&pub_key)?,
            database_url: std::env::var("DATABASE_URL")?,
            listen_port: std::env::var("LISTEN_PORT")?.parse()?,
        })
    }

    fn authorization(&self) -> String {
        let engine = engine::GeneralPurpose::new(&alphabet::STANDARD, engine::general_purpose::PAD);
        let auth = format!("{}:{}", self.discord_client_id, self.discord_client_secret);

        engine.encode(auth)
    }
}

#[derive(Clone)]
struct AppState {
    config: Config,
    pg_pool: PgPool,
}

impl FromRef<AppState> for PgPool {
    fn from_ref(app_state: &AppState) -> Self {
        app_state.pg_pool.clone()
    }
}

impl FromRef<AppState> for VerifyingKey {
    fn from_ref(app_state: &AppState) -> Self {
        app_state.config.discord_pub_key
    }
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    let config = Config::configure()?;
    let tracer = opentelemetry_jaeger::new_agent_pipeline()
        .with_service_name("god_replacement_product")
        .install_simple()?;
    let telemetry = tracing_opentelemetry::layer().with_tracer(tracer);
    let (loki_layer, loki_task) = tracing_loki::builder()
        .label("host", "mine")?
        .extra_field("pid", format!("{}", process::id()))?
        .build_url(url::Url::parse("http://127.0.0.1:3100").unwrap())?;

    tracing_subscriber::registry()
        .with(telemetry.with_filter(LevelFilter::DEBUG))
        .with(
            tracing_subscriber::fmt::layer()
                .and_then(loki_layer)
                .with_filter(
                    tracing_subscriber::EnvFilter::try_from_default_env().unwrap_or_else(|_| {
                        "god_replacement_product=debug,tower_http=debug".into()
                    }),
                ),
        )
        .init();

    let port = config.listen_port;

    let pg_pool = PgPoolOptions::new()
        .max_connections(5)
        .connect(&config.database_url)
        .await?;

    sqlx::migrate!().run(&pg_pool).await?;

    register_commands(config.discord_client_id.to_owned(), config.authorization()).await?;
    let state = AppState { config, pg_pool };

    let app = Router::new()
        .route("/api/discord/interactions/", post(post_interaction))
        .with_state(state)
        .layer(TraceLayer::new_for_http());

    let addr = SocketAddr::from(([127, 0, 0, 1], port));
    tracing::debug!("listening on {}", addr);
    tokio::spawn(loki_task);

    axum::Server::bind(&addr)
        .serve(app.into_make_service())
        .with_graceful_shutdown(shutdown_signal())
        .await?;

    tracing::info!("server down");

    opentelemetry::global::shutdown_tracer_provider();

    Ok(())
}

async fn shutdown_signal() {
    let ctrl_c = async {
        signal::ctrl_c()
            .await
            .expect("failed to install Ctrl+C handler");
    };

    #[cfg(unix)]
    let terminate = async {
        signal::unix::signal(signal::unix::SignalKind::terminate())
            .expect("failed to install signal handler")
            .recv()
            .await;
    };

    #[cfg(not(unix))]
    let terminate = std::future::pending::<()>();

    tokio::select! {
        _ = ctrl_c => {},
        _ = terminate => {},
    }

    tracing::info!("signal received, starting graceful shutdown");
}

type InteractionResult = Result<(StatusCode, Json<InteractionResponse>), (StatusCode, String)>;

async fn post_interaction(
    State(pg_pool): State<PgPool>,
    ExtractInteraction(interaction): ExtractInteraction,
) -> InteractionResult {
    match interaction.kind {
        InteractionType::Ping => {
            let pong = InteractionResponse {
                kind: InteractionResponseType::Pong,
                data: None,
            };

            Ok((StatusCode::OK, Json(pong)))
        }
        InteractionType::ApplicationCommand => {
            let author_id = interaction.author_id();
            let Some(InteractionData::ApplicationCommand(data)) = interaction.data.clone() else {
                return not_found();
            };
            let command_input_data = CommandInputData::from(*data.clone());
            tracing::debug!(command_name = ?data.name, "started processing command");
            let result = match &*data.name {
                SetFactCommand::NAME => {
                    let Ok(command_data) = SetFactCommand::from_interaction(command_input_data) else {
                        return Err((StatusCode::BAD_REQUEST, format!("invalid {0} command.", SetFactCommand::NAME)));
                    };
                    let Some(author_id) = author_id else {
                        return Err((StatusCode::BAD_REQUEST, format!("{0} requires a user.", SetFactCommand::NAME)));
                    };
                    match set_fact(
                        interaction.id,
                        interaction.channel_id,
                        author_id,
                        command_data,
                        &pg_pool,
                    )
                    .await
                    {
                        Ok(response) => Ok((StatusCode::OK, Json(response))),
                        Err(err) => Err(err),
                    }
                }
                GetFactCommand::NAME => {
                    let Ok(command_data) = GetFactCommand::from_interaction(command_input_data) else {
                        return Err((StatusCode::BAD_REQUEST, format!("invalid {0} command.", GetFactCommand::NAME)));
                    };
                    let Some(author_id) = author_id else {
                        return Err((StatusCode::BAD_REQUEST, format!("{0} requires a user.", GetFactCommand::NAME)));
                    };

                    match get_fact(interaction.channel_id, author_id, command_data, &pg_pool).await
                    {
                        Ok(response) => Ok((StatusCode::OK, Json(response))),
                        Err(err) => Err(err),
                    }
                }
                _ => not_found(),
            };

            tracing::debug!(command_name = ?data.name, "finished processing command");
            result
        }
        _ => not_found(),
    }
}

fn not_found() -> InteractionResult {
    Err((
        StatusCode::NOT_FOUND,
        "requested interaction not found".to_string(),
    ))
}
